(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: pluginTraceMcdc.ml                                            *)
(* authors: Adrien Jonquet, Mathias Bourgoin, A. Darrasse, P. Wang     *)
(* licence: CeCILL-B                                                   *)
(***********************************************************************)

open TraceMcdc


module Events = struct
  (* N.B. by phil (2012, March): Events can't use maps or sets! *)
  type 'a t = (int * 'a) list
  let empty = []
  let add k x l = (k, x) :: l
  let iter f = List.iter (fun (k, x) -> f k x)
  let find e l =
    let rec find res = function
      | (a,b) :: tl ->
          if a = e  then 
            find (b::res) tl
          else
            find res tl
      | [] -> res
    in find [] l
  let mem = List.mem_assoc
end


type events = Covevent.coverage_event Events.t

(* (* TODO: use maps rather than lists. DONE: by phil in March 2012 *) *)
(* let mcdc_trace = ref ([] : ((int * string) * (mcdc list * events * events)) list)  *)

module M = Map.Make(struct type t = int * string let compare (a,_) (b,_) = compare a b end)
let mcdc_trace = ref (M.empty : (mcdc list * events * events) M.t)

open Instructions

let nb_vectors = ref 0
let nb_merged = ref 0

let vector_completed vector = vector.vec_dec.nb_dec > 0

let rec merge_vector (m:TraceMcdc.mcdc) (vector:TraceMcdc.vector) : TraceMcdc.vector list -> unit = function
    [] -> m.vectors <- vector :: List.tl m.vectors 
  | v :: q ->
      let b = ref true in
      for i = 0 to Array.length v.vec_conds - 1 do
        assert(vector.vec_conds.(i).cond = v.vec_conds.(i).cond);
        match vector.vec_conds.(i).cond_value, v.vec_conds.(i).cond_value with
        | Unknown, Unknown -> ()
        | Bool true, Bool true -> ()
        | Bool false, Bool false -> ()
        | _ -> b := false
      done;
      if (!b) && (vector.vec_dec.dec_value = v.vec_dec.dec_value) then
        begin
          v.vec_dec.nb_dec <- v.vec_dec.nb_dec + 1;
          m.vectors <- List.tl m.vectors
        end
      else
        merge_vector m vector q


let incr_vec vector =
  vector.nb_vec <- vector.nb_vec + 1;
  vector.vec_dec.nb_dec <- vector.vec_dec.nb_dec + 1

let record_dec dec v m =
  let new_vector b = {
    nb_vec = 0;
    vec_dec = {
      dec_value = Bool b;
      nb_dec = 0;
      dec = m.model.mod_dec
    };
    vec_conds = [||]
  } in
  if m.model.mod_conds = [] then begin
    if m.vectors = [] then m.vectors <- [ new_vector true; new_vector false ];
    incr_vec (List.nth m.vectors (if v then 0 else 1))
  end else
    let vector = List.hd m.vectors in
    vector.vec_dec <- { dec_value = Bool v; nb_dec = 1; dec = m.model.mod_dec };
  if List.length m.vectors > 1 then
    merge_vector m (List.hd m.vectors) (List.tl m.vectors)

let record_cond (cond:Covevent.coverage_event) (v:bool) (m:TraceMcdc.mcdc) : unit =
  let same_loc (cond:Covevent.coverage_event) (cond':Covevent.coverage_event) : bool =
    (cond'.Covevent.covev_loc:Location.t) = (cond.Covevent.covev_loc:Location.t)
  in
  if List.mem cond (m.model.mod_conds:Covevent.coverage_event list) then begin
    if (m.vectors = []) || (cond = List.hd m.model.mod_conds) then
      let new_vec_cond i =
        let eq = same_loc cond (List.nth m.model.mod_conds i) in
        { cond_value = if eq then Bool v else Unknown;
          nb_cond = if eq then 1 else 0;
          cond = List.nth m.model.mod_conds i } in
      m.vectors <- {
        nb_vec = 0;
        vec_dec = {
          dec_value = Unknown;
          nb_dec = 0;
          dec = m.model.mod_dec
        };
        vec_conds = Array.init (List.length m.model.mod_conds) new_vec_cond
      } :: m.vectors
    else
      let vector = List.hd m.vectors in
      for i=0 to (Array.length vector.vec_conds)-1 do
        let vec_cond = vector.vec_conds.(i) in
        if same_loc cond vec_cond.cond then begin
          vec_cond.nb_cond <- vec_cond.nb_cond + 1;
          vec_cond.cond_value <- Bool v
        end
      done
  end

let find_cov_from_addr pc l =
  Events.find pc l

let cond_value vm i = 
  let b =
    match i with
      | BOOLNOT -> vm.Vm.accumulator <> Value.Int 0
      | BRANCHIF n -> vm.Vm.accumulator <> Value.Int 0
      | BRANCHIFNOT n -> vm.Vm.accumulator <> Value.Int 0
      | RETURN n -> vm.Vm.accumulator <> Value.Int 0
      | ACC0 -> Vm.peek vm 0 <> Value.Int 0
      | ACC1 -> Vm.peek vm 1 <> Value.Int 0
      | ACC2 -> Vm.peek vm 2 <> Value.Int 0
      | ACC3 -> Vm.peek vm 3 <> Value.Int 0
      | ACC4 -> Vm.peek vm 4 <> Value.Int 0
      | ACC5 -> Vm.peek vm 5 <> Value.Int 0
      | ACC6 -> Vm.peek vm 6 <> Value.Int 0
      | ACC7 -> Vm.peek vm 7 <> Value.Int 0
      | ACC n -> Vm.peek vm n <> Value.Int 0
      | CONST0 -> false
      | CONST1 -> true
      | i -> vm.Vm.accumulator <> Value.Int 0
  in
    if false then (* DEBUG STUFF *)
      begin
        Printf.eprintf "[%s ==> %b] --- %s\n%!"
          (Instructions.string_of_instructions [||] i) 
          b
          (Utils.string_of_value (List.hd vm.Vm.stack)) 
      end;
    b

let cond_value vm i =
  let b =
    match i with
      | CONST0 -> false
      | CONST1 -> true
      | _ ->
          (* manual (bad?) optimization versus simplicity... *)
          vm.Vm.accumulator <> Value.Int 0
  in
    if false then (* DEBUG STUFF *)
      begin
        Printf.eprintf "[%s ==> %b] --- %s\n%!"
          (Instructions.string_of_instructions [||] i) 
          b
          (Utils.string_of_value (List.hd vm.Vm.stack)) 
      end;
    b


(* let cond_value_b vm i = *)
(*   match i with  *)
(*     | CONST0 -> Some false  *)
(*     | CONST1 -> Some true  *)
(* (\*     | RETURN _ ->   *\) *)
(* (\*         if vm.Vm.accumulator <> List.hd vm.Vm.stack then  *\) *)
(* (\*           match List.hd vm.Vm.accumulator with *\) *)
(* (\*             | Value.Int 0 -> Some false *\) *)
(* (\*             | Value.Int 1 -> Some true *\) *)
(* (\*             | _ -> assert false *\) *)
(* (\*         else *\) *)
(* (\*           None   *\)    *)

(*     | RETURN _  *)
(*     | BRANCHIFNOT _  *)
(*     | PUSHACC2 *)
(*     | _ ->  *)
(*         let r =  *)
(*           match vm.Vm.accumulator with  *)
(*             | Value.Int 0 -> Some false *)
(*             | Value.Int 1 -> Some true *)
(*             | _ -> *)
(*                 None *)
(*         in *)
(*           Printf.printf "[%s] --- accu=%s stacktop=%s\n%!" *)
(*             (Instructions.string_of_instructions [||] i)  *)
(*             (Utils.string_of_value vm.Vm.accumulator) *)
(*             (Utils.string_of_value (List.hd vm.Vm.stack)); *)
(*           r *)



let string_of_location (l:Location.t) =
  let open Location in
  let open Lexing in
    match l with { loc_start = { pos_lnum = sl ; pos_cnum = sc } ;
                   loc_end = { pos_lnum = el ; pos_cnum = ec } } ->
      Printf.sprintf "%3d--%4d-%4d" sl sc ec
      

(* record the boolean vectors for MC/DC *)
let record_mcdc =
  let last_instr = ref STOP in
    fun
      (vm:Vm.virtual_machine)
      ((mcdc_trace:TraceMcdc.mcdc list),
      (cond_list:(Covevent.coverage_event Events.t)),
      (dec_list:(Covevent.coverage_event Events.t)))
      (instr:Instructions.instruction)
      ->
        begin
          let cel : Covevent.coverage_event list =
            find_cov_from_addr (vm.Vm.code_pointer * 4) cond_list
          in
            List.iter (fun cond ->
              match (instr, !last_instr) with
                  (RETURN _, BRANCHIF _) | (RETURN _, BRANCHIFNOT _) -> ()
                | (_, BRANCHIFNOT _) -> ()
                | (_, BRANCHIF _) -> ()
                | _ -> 
                    match cond_value vm instr with
                      | v ->
                          (* let v : bool = cond_value vm instr in *)
                          if false then (* DEBUG STUFF *)
                            (Printf.eprintf "[%s] instr/last_instr [%s@%d--%s] ==> %b\n%!" 
                                (string_of_location cond.Covevent.covev_loc)
                                (Instructions.string_of_instructions [||] instr)
                                vm.Vm.code_pointer
                                (Instructions.string_of_instructions [||] !last_instr)
                                v;
                            );
                          List.iter (record_cond cond v) mcdc_trace
            ) cel;
        end;
        let cel : Covevent.coverage_event list =
          find_cov_from_addr (vm.Vm.code_pointer * 4) dec_list 
        in
          begin
            List.iter (fun dec -> 
              let v = vm.Vm.accumulator <> Value.Int 0 in
              let m = List.find (fun m -> m.model.mod_dec = dec) mcdc_trace in
                record_dec dec v m
            ) cel
          end;
          last_instr := instr

let remove_dups =
  let filter lu event =
    let test a = CoverUtils.get_loc a = CoverUtils.get_loc event in
    if List.exists test lu then lu else (event :: lu) in
  List.fold_left filter []

(* get the list of decisions and the list of conditions *)
let first_pass =
  let aux (min_pos, mcdc_l, cond_l, dec_l) cov =
    let new_min_pos = min min_pos cov.Covevent.covev_pos in
    match cov.Covevent.covev_status with
    | Covevent.Coverage_event_decision ->
        let vect = {
          model = {mod_dec = cov; mod_conds = []};
          vectors = []
        } in
        (new_min_pos, (vect::mcdc_l), cond_l, Events.add cov.Covevent.covev_pos cov dec_l)
    | Covevent.Coverage_event_condition _ ->
        (new_min_pos, mcdc_l, Events.add cov.Covevent.covev_pos cov cond_l, dec_l)
    | _ -> (new_min_pos, mcdc_l, cond_l, dec_l) in
    List.fold_left aux (max_int, [], Events.empty, Events.empty)

let rec second_pass mcdc_list =
  let aux cond dec t =
    if t.model.mod_dec.Covevent.covev_loc = dec.Covevent.covev_loc then
      t.model.mod_conds <- cond :: t.model.mod_conds in
  let populate_vector_model cond dec = List.iter (aux cond dec) mcdc_list in
  let populate_mcdc _ cond = match cond.Covevent.covev_status with
    Covevent.Coverage_event_condition dec -> populate_vector_model cond dec
  | _ -> assert false in
  Events.iter populate_mcdc

(* build the boolean vectors for MC/DC *)
let build_mcdc_vectors data module_name =
  (* get the list of modules *)
  let ci = open_in (module_name ^ ".cmo.cover") in
  let (orig, ev_list) = Marshal.from_channel ci in
  close_in ci;
(*  if ev_list = [] then failwith ("no coverage_events for module " ^ module_name); *)
  List.iter (CoverUtils.relocate_event orig) ev_list;
  let min_pos, mcdc_list, cond_list, dec_list = first_pass (remove_dups ev_list) in
  second_pass mcdc_list cond_list;
  (min_pos, mcdc_list, cond_list, dec_list)

let mcdc_trace_enabled = ref false
let mcdc_trace_name = ref ""

let init data =
(* TODO: move the vector building in compile-time. *)
  let init_module module_name =
    let min_pos, mcdc_list, cond_list, dec_list = build_mcdc_vectors data module_name in
    ((min_pos, module_name), (mcdc_list, cond_list, dec_list)) in
  if !mcdc_trace_enabled then
    let module_names = CoverUtils.get_module_names () in
      mcdc_trace := List.fold_left (fun r e -> let a,b = (init_module e) in M.add a b r) M.empty module_names


exception Break
let step vm instruction =
  let rec step_module = function
    ((min_pos, _), mcdc_trace_list) :: ls ->
      if vm.Vm.code_pointer * 4 >= min_pos then
        record_mcdc vm mcdc_trace_list instruction
      else step_module ls
  | [] -> () in
  (* record the current obj instruction for source coverage *)
  (* FIXME: apply step_module only for the module including the current instruction. *)
  (* if !mcdc_trace_enabled then step_module !mcdc_trace *)
  let step_module (min_pos, _) mcdc_trace_list = 
    if vm.Vm.code_pointer * 4 >= min_pos then
      (record_mcdc vm mcdc_trace_list instruction; raise Break)
    else 
      ()
  in
    try if !mcdc_trace_enabled then M.iter step_module !mcdc_trace with Break -> ()


let extract_trace map =
  M.fold (fun k (trace, _, _) r -> (k,trace) :: r)  map []


let finalise () =
  if !mcdc_trace_enabled then
    (* mcdc trace generation *)
    try begin
      let co = open_out !mcdc_trace_name in
        Marshal.to_channel co (extract_trace !mcdc_trace) [];
      close_out co
    end with
    | Sys_error s -> failwith ("pluginTraceMcdc.ml : "^s)

let _ =
  Plugin.register_plugin init step finalise
    ("-mcdc", Arg.String (fun s -> mcdc_trace_enabled := true; mcdc_trace_name := s),
     "file write mcdc trace in file")
